from bbcpy.functions.statistics import cov
import numpy as np
from scipy.linalg import eigh
from sklearn.base import ClassifierMixin, TransformerMixin, BaseEstimator
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.multiclass import unique_labels
from pyriemann.utils import covariance
from pyriemann.stats import mean_covariance


class Wishart(BaseEstimator, ClassifierMixin, TransformerMixin):
    def __init__(self, estimator=covariance._scm, excllev=None, autobias=False, geomean=False, covupdate=False):
        self.estimator = estimator
        self.excllev = excllev
        self.autobias = autobias
        self.geomean = geomean
        self.covupdate = covupdate

    def fit(self, X, y=None):
        """

        :param X: epoched data
        :param y: class-label
        :return:
        """
        if y is not None:
            X.y = y
            X.mrk.y = y
        self.classes_ = unique_labels(y)
        Sigma_trial = cov(X, target='trial')
        if self.geomean:
            self.Sigma_ = [mean_covariance(Sigma_trial[y == 0]), mean_covariance(Sigma_trial[y == 1])]
        elif self.excllev is not None:
            covtr = np.trace(np.linalg.pinv(cov(X, target='all', estimator=self.estimator)) @ Sigma_trial, axis1=1,
                             axis2=2) / X.shape[1]
            sel_tr = covtr <= self.excllev
            self.Sigma_ = cov(X[sel_tr], target='class', estimator=self.estimator)
        else:
            self.Sigma_ = cov(X, target='class', estimator=self.estimator)
        # self.SigmaA_ = cov(X, target='all', estimator=self.estimator)
        self.Sigmainv_ = [np.linalg.pinv(self.Sigma_[0]), np.linalg.pinv(self.Sigma_[1])]

        if self.autobias:
            wx = (np.trace(self.Sigmainv_[0] @ Sigma_trial, axis1=1, axis2=2) - np.trace(
                self.Sigmainv_[1] @ Sigma_trial, axis1=1, axis2=2))
            self.b_ = - wx.mean()
        else:
            self.b_ = (np.log(np.linalg.det(self.Sigma_[0])) - np.log(np.linalg.det(self.Sigma_[1])))
            self.b_v_ = np.concatenate([(np.log(np.linalg.det(self.Sigma_[0])), np.log(np.linalg.det(self.Sigma_[1])))])
        return self

    def transform(self, X):
        """
        :param X: ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series
        :param y: ndarray, shape (n_trials, 2)
            Marker positions and marker class.
        :return: out : ndarray, shape (n_matrices, n_csp, n_time)
            transformed
        """
        check_is_fitted(self)
        n = X.shape[2]
        Sigma = cov(X, target='trial')
        wx = np.concatenate([(np.trace(self.Sigmainv_[0] @ Sigma, axis1=1, axis2=2), np.trace(self.Sigmainv_[1] @ Sigma,
                                                                                              axis1=1, axis2=2))])
        return (n - 1) * wx.T + n * self.b_v_

    def decision_function(self, X):
        """
        Predict confidence scores for samples.
        The confidence score for a sample is proportional to the signed
        distance of that sample to the hyperplane.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the confidence scores.
        Returns
        -------
        scores : ndarray of shape (n_samples,) or (n_samples, n_classes)
            Confidence scores per `(n_samples, n_classes)` combination. In the
            binary case, confidence score for `self.classes_[1]` where >0 means
            this class would be predicted.
        """
        check_is_fitted(self)
        n = X.shape[2]
        Sigma = cov(X, target='trial')
        wx = (np.trace(self.Sigmainv_[0] @ Sigma, axis1=1, axis2=2) - np.trace(self.Sigmainv_[1] @ Sigma, axis1=1,
                                                                               axis2=2))
        y = (n - 1) * wx + n * self.b_
        if self.covupdate:
            # deltaS = cov(X, target='all', estimator=self.estimator) - self.SigmaA_
            # Sigma = Sigma - deltaS
            y = y - (n - 1) * np.trace(
                (self.Sigmainv_[0] - self.Sigmainv_[1]) @ np.mean(Sigma, axis=0),
                axis1=0, axis2=1)

        return y

    def predict(self, X):
        """
        Predict class labels for samples in X.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the predictions.
        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(int)
        else:
            indices = scores.argmax(axis=1)
        return self.classes_[indices]


class Miklodyfire(BaseEstimator, ClassifierMixin, TransformerMixin):
    def __init__(self, n_cmps=6, estimator=covariance._scm, excllev=None, autobias=False):
        self.n_cmps = n_cmps
        self.estimator = estimator
        self.excllev = excllev
        self.autobias = autobias

    def fit(self, X, y=None):
        """

        :param X: epoched data
        :param y: class-label
        :return:
        """
        if y is not None:
            X.y = y
            X.mrk.y = y
        self.classes_ = unique_labels(y)
        Sigma_trial = cov(X, target='trial')
        if self.excllev is not None:
            covtr = np.trace(np.linalg.pinv(cov(X, target='all', estimator=self.estimator)) @ Sigma_trial, axis1=1,
                             axis2=2) / X.shape[1]
            sel_tr = covtr <= self.excllev
            self.Sigma_ = cov(X[sel_tr], target='class', estimator=self.estimator)
        else:
            self.Sigma_ = cov(X, target='class', estimator=self.estimator)
        # d, W = eigh(np.linalg.pinv(self.Sigma_[0])-np.linalg.pinv(self.Sigma_[1]),
        ##                      np.linalg.pinv(self.Sigma_[0])+np.linalg.pinv(self.Sigma_[1])/2)
        d, W = eigh(self.Sigma_[0] - self.Sigma_[1], (self.Sigma_[0] + self.Sigma_[1]) / 2)
        # selected_csps = np.flipud(np.argsort(np.maximum(d, 1 - d)))[:self.n_cmps]
        selected_csps = np.flipud(np.argsort(np.abs(d)))[:self.n_cmps]
        self.W = W[:, selected_csps]
        self.d = d[selected_csps]
        self.Sigma_ = [self.W.T @ self.Sigma_[0] @ self.W, self.W.T @ self.Sigma_[1] @ self.W]
        # self.Sigmainv_ = W[:, selected_csps] @ np.diag(d[selected_csps]) @ W[:, selected_csps].T
        # self.Sigmainv_ = [W[:, selected_csps].T @ np.linalg.pinv(self.Sigma_[0])  W[:, selected_csps],
        #     W[:, selected_csps].T @ np.linalg.pinv(self.Sigma_[1]) @ W[:, selected_csps]]
        self.Sigmainv_ = [np.linalg.pinv(self.Sigma_[0]), np.linalg.pinv(self.Sigma_[1])]

        if self.autobias:
            Sigma = self.W.T @ Sigma_trial @ self.W
            wx = (np.trace(self.Sigmainv_[0] @ Sigma, axis1=1, axis2=2) - np.trace(self.Sigmainv_[1] @ Sigma,
                                                                                   axis1=1, axis2=2))
            self.b_ = - wx.mean()
        else:

            self.b_ = (np.log(np.linalg.det(self.Sigma_[0])) - np.log(np.linalg.det(self.Sigma_[1])))
            self.b_v_ = np.concatenate([(np.log(np.linalg.det(self.Sigma_[0])), np.log(np.linalg.det(self.Sigma_[1])))])
        return self

    def transform(self, X):
        """
        :param X: ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series
        :param y: ndarray, shape (n_trials, 2)
            Marker positions and marker class.
        :return: out : ndarray, shape (n_matrices, n_csp, n_time)
            transformed
        """
        check_is_fitted(self)
        n = X.shape[2]
        Sigma = self.W.T @ cov(X, target='trial') @ self.W

        wx = np.concatenate([(np.trace(self.Sigmainv_[0] @ Sigma, axis1=1, axis2=2), np.trace(self.Sigmainv_[1] @ Sigma,
                                                                                              axis1=1, axis2=2))])
        return (n - 1) * wx.T + n * self.b_v_

    def decision_function(self, X):
        """
        Predict confidence scores for samples.
        The confidence score for a sample is proportional to the signed
        distance of that sample to the hyperplane.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the confidence scores.
        Returns
        -------
        scores : ndarray of shape (n_samples,) or (n_samples, n_classes)
            Confidence scores per `(n_samples, n_classes)` combination. In the
            binary case, confidence score for `self.classes_[1]` where >0 means
            this class would be predicted.
        """
        check_is_fitted(self)
        n = X.shape[2]
        Sigma = self.W.T @ cov(X, target='trial') @ self.W
        wx = (np.trace(self.Sigmainv_[0] @ Sigma, axis1=1, axis2=2) - np.trace(self.Sigmainv_[1] @ Sigma,
                                                                               axis1=1, axis2=2))
        return (n - 1) * wx + n * self.b_

    def predict(self, X):
        """
        Predict class labels for samples in X.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the predictions.
        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(int)
        else:
            indices = scores.argmax(axis=1)
        return self.classes_[indices]


class Miklodyfire2(BaseEstimator, ClassifierMixin, TransformerMixin):
    def __init__(self, n_cmps=6, estimator=covariance._scm, excllev=None, autobias=False):
        self.n_cmps = n_cmps
        self.estimator = estimator
        self.excllev = excllev
        self.autobias = autobias

    def fit(self, X, y=None):
        """

        :param X: epoched data
        :param y: class-label
        :return:
        """
        if y is not None:
            X.y = y
            X.mrk.y = y
        self.classes_ = unique_labels(y)
        n = X.shape[2]
        Sigma_trial = cov(X, target='trial')
        if self.excllev is not None:
            covtr = np.trace(np.linalg.pinv(cov(X, target='all', estimator=self.estimator)) @ Sigma_trial, axis1=1,
                             axis2=2) / X.shape[1]
            sel_tr = covtr <= self.excllev
            self.Sigma_ = cov(X[sel_tr], target='class', estimator=self.estimator)
        else:
            self.Sigma_ = cov(X, target='class', estimator=self.estimator)
        d, W = eigh(np.linalg.pinv(self.Sigma_[0]) - np.linalg.pinv(self.Sigma_[1]))
        # d, W = eigh(self.Sigma_[0] - self.Sigma_[1], (self.Sigma_[0] + self.Sigma_[1]) / 2)
        selected_csps = np.flipud(np.argsort(np.abs(d)))[:self.n_cmps]
        self.W = W[:, selected_csps]
        self.d = d[selected_csps]
        self.Sigmainv_ = self.W @ np.diag(self.d) @ self.W.T

        if self.autobias:
            wx = np.trace(self.Sigmainv_ @ Sigma_trial, axis1=1, axis2=2)
            self.b_ = - (n - 1) / n * wx.mean()
            # out = (n - 1) * wx - n * wx.mean()
            # self.b_ = np.mean((n-1)/n * wx - 1/n * out)
        else:
            # self.b_ = - np.log(np.prod(self.d))
            # self.b_ = np.log(np.linalg.det(np.eye(X.shape[1])-self.Sigma_[0] @ self.Sigmainv_))
            # self.b_ = (np.log(np.linalg.det(self.Sigma_[0])) - np.log(np.linalg.det(self.Sigma_[1])))
            self.b_ = (np.log(np.linalg.det(self.W.T @ self.Sigma_[0] @ self.W)) -
                       np.log(np.linalg.det(self.W.T @ self.Sigma_[1] @ self.W)))
        return self

    def transform(self, X):
        """
        :param X: ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series
        :param y: ndarray, shape (n_trials, 2)
            Marker positions and marker class.
        :return: out : ndarray, shape (n_matrices, n_csp, n_time)
            transformed
        """
        return X

    def decision_function(self, X):
        """
        Predict confidence scores for samples.
        The confidence score for a sample is proportional to the signed
        distance of that sample to the hyperplane.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the confidence scores.
        Returns
        -------
        scores : ndarray of shape (n_samples,) or (n_samples, n_classes)
            Confidence scores per `(n_samples, n_classes)` combination. In the
            binary case, confidence score for `self.classes_[1]` where >0 means
            this class would be predicted.
        """
        check_is_fitted(self)
        n = X.shape[2]
        Sigma = cov(X, target='trial')
        wx = np.trace(self.Sigmainv_ @ Sigma, axis1=1, axis2=2)
        return (n - 1) * wx + n * self.b_

    def predict(self, X):
        """
        Predict class labels for samples in X.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the predictions.
        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(int)
        else:
            indices = scores.argmax(axis=1)
        return self.classes_[indices]


class Miklodyfire3(BaseEstimator, ClassifierMixin, TransformerMixin):
    def __init__(self, n_cmps=6, estimator=covariance._scm, excllev=None, autobias=False):
        self.n_cmps = n_cmps
        self.estimator = estimator
        self.excllev = excllev
        self.autobias = autobias

    def fit(self, X, y=None):
        """

        :param X: epoched data
        :param y: class-label
        :return:
        """
        if y is not None:
            X.y = y
            X.mrk.y = y
        self.classes_ = unique_labels(y)
        n = X.shape[2]
        Sigma_trial = cov(X, target='trial')
        if self.excllev is not None:
            covtr = np.trace(np.linalg.pinv(cov(X, target='all', estimator=self.estimator)) @ Sigma_trial, axis1=1,
                             axis2=2) / X.shape[1]
            sel_tr = covtr <= self.excllev
            self.Sigma_ = cov(X[sel_tr], target='class', estimator=self.estimator)
        else:
            self.Sigma_ = cov(X, target='class', estimator=self.estimator)
        d, W = eigh(cov(X, target='all', estimator=self.estimator))
        d1 = np.diag(W.T @ self.Sigma_[0] @ W)
        d2 = np.diag(W.T @ self.Sigma_[1] @ W)
        selected_csps = np.flipud(np.argsort(np.abs(np.log(d1) - np.log(d2))))[:self.n_cmps]
        self.W = W[:, selected_csps]
        # self.d = d[selected_csps]
        self.Sigmainv_ = self.W @ np.diag(1 / d1[selected_csps]) @ self.W.T - self.W @ np.diag(
            1 / d2[selected_csps]) @ self.W.T

        if self.autobias:
            wx = np.trace(self.Sigmainv_ @ Sigma_trial, axis1=1, axis2=2)
            self.b_ = - (n - 1) / n * wx.mean()
            # out = (n - 1) * wx - n * wx.mean()
            # self.b_ = np.mean((n-1)/n * wx - 1/n * out)
        else:
            self.b_ = - np.log(np.prod(np.concatenate([d1[selected_csps], 1 / d2[selected_csps]])))
            # self.b_ = - np.log(np.prod(np.concatenate([self.W @ np.diag(d1[selected_csps]) @ self.W.T,
            #                                           self.W @ np.diag(1/d2[selected_csps]) @ self.W.T])))
            # self.b_ = np.log(np.linalg.det(np.eye(X.shape[1])-self.Sigma_[0] @ self.Sigmainv_))
            # self.b_ = (np.log(np.linalg.det(self.Sigma_[0])) - np.log(np.linalg.det(self.Sigma_[1])))
            # self.b_ = (np.log(np.linalg.det(self.W.T @ self.Sigma_[0] @ self.W)) -
            #           np.log(np.linalg.det(self.W.T @ self.Sigma_[1] @ self.W)))
        return self

    def transform(self, X):
        """
        :param X: ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series
        :param y: ndarray, shape (n_trials, 2)
            Marker positions and marker class.
        :return: out : ndarray, shape (n_matrices, n_csp, n_time)
            transformed
        """
        return X

    def decision_function(self, X):
        """
        Predict confidence scores for samples.
        The confidence score for a sample is proportional to the signed
        distance of that sample to the hyperplane.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the confidence scores.
        Returns
        -------
        scores : ndarray of shape (n_samples,) or (n_samples, n_classes)
            Confidence scores per `(n_samples, n_classes)` combination. In the
            binary case, confidence score for `self.classes_[1]` where >0 means
            this class would be predicted.
        """
        check_is_fitted(self)
        n = X.shape[2]
        Sigma = cov(X, target='trial')
        wx = np.trace(self.Sigmainv_ @ Sigma, axis1=1, axis2=2)
        return (n - 1) * wx + n * self.b_

    def predict(self, X):
        """
        Predict class labels for samples in X.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the predictions.
        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(int)
        else:
            indices = scores.argmax(axis=1)
        return self.classes_[indices]



import os, sys

sys.path.append(os.getcwd())
import scipy.io as sio
import demos.vitalbci.vitalbci_funcs as vitalbci_funcs
import numpy as np

if __name__ == '__main__':
    if len(sys.argv) > 1:
        if sys.argv[1] == 'None':
            excllev = None
            exclstr = 'None'
        else:
            excllev = float(sys.argv[1])
            exclstr = '%i' % (excllev*10)
    else:
        excllev = 1.5
    if len(sys.argv) > 2:
        estimator = sys.argv[2]
    else:
        estimator = 'lwf'
    print(excllev)
    print(estimator)

vp_fb_parameters = sio.loadmat('/home/bbci/data/results/studies/vitalbci_season1/vp_fb_parameters.mat')
performance = sio.loadmat('/home/bbci/data/results/studies/vitalbci_season1/performance.mat')
subdir_list = performance['subdir_list']
ivals = vp_fb_parameters['ival']
bands = vp_fb_parameters['band']
classes = vp_fb_parameters['classes']
nCSPs = vp_fb_parameters['nCSP']
Nsubj = len(subdir_list)
cvresultsAlpha = [None] * Nsubj
cvresultsBeta = [None] * Nsubj
cvresultsTheta = [None] * Nsubj
cvresultsBroad = [None] * Nsubj

for scoring in [None, 'roc_auc']:
    for i, fname in enumerate(subdir_list):
        fname = fname[0][0]
        subjname = fname.split('_')[0]
        fullfile = '/home/bbci/data/bbciMat/' + fname + '/imag_arrow' + subjname + '.mat'  # _250Hz option
        ival = ivals[i].astype(int)
        band = bands[i]
        nCSP = nCSPs[i][0]
        classi = [str(classes[i][0][0][0][0]), str(classes[i][0][0][1][0])]
        if band[0] < 13:
            cvresultsTheta[i] = vitalbci_funcs.load_and_clfy_Miklodyfire(fullfile, ival, band / 2, classi, nCSP,
                                                                         scoring=scoring, estimator=estimator,
                                                                         excllev=excllev)
            cvresultsAlpha[i] = vitalbci_funcs.load_and_clfy_Miklodyfire(fullfile, ival, band, classi, nCSP,
                                                                         scoring=scoring, estimator=estimator,
                                                                         excllev=excllev)
            if band[1] > 22.5:
                bandmax = 45
            else:
                bandmax = 2 * band[1]
            cvresultsBeta[i] = vitalbci_funcs.load_and_clfy_Miklodyfire(fullfile, ival, [band[0] * 2, bandmax], classi,
                                                                        nCSP, scoring=scoring, estimator=estimator,
                                                                        excllev=excllev)
            cvresultsBroad[i] = vitalbci_funcs.load_and_clfy_Miklodyfire(fullfile, ival, [band[0], bandmax], classi,
                                                                         nCSP, scoring=scoring, estimator=estimator,
                                                                         excllev=excllev)
        if band[0] >= 13:
            cvresultsTheta[i] = vitalbci_funcs.load_and_clfy_Miklodyfire(fullfile, ival, band / 4, classi, nCSP,
                                                                         scoring=scoring, estimator=estimator,
                                                                         excllev=excllev)
            cvresultsAlpha[i] = vitalbci_funcs.load_and_clfy_Miklodyfire(fullfile, ival, band / 2, classi, nCSP,
                                                                         scoring=scoring, estimator=estimator,
                                                                         excllev=excllev)
            cvresultsBeta[i] = vitalbci_funcs.load_and_clfy_Miklodyfire(fullfile, ival, band, classi, nCSP,
                                                                        scoring=scoring, estimator=estimator,
                                                                        excllev=excllev)
            cvresultsBroad[i] = vitalbci_funcs.load_and_clfy_Miklodyfire(fullfile, ival, band * [0.5, 1], classi, nCSP,
                                                                         scoring=scoring, estimator=estimator,
                                                                         excllev=excllev)
        maxres = np.max([np.max(np.mean(cvresultsAlpha[i], axis=1)), np.max(np.mean(cvresultsBeta[i], axis=1)),
                         np.max(np.mean(cvresultsTheta[i], axis=1)), np.max(np.mean(cvresultsBroad[i], axis=1))])
        print('Subject finished, maximum accuracy was %2.1f%% while original was %2.1f%%' % (100 * maxres,
                                                                                             performance['acc_c'][0][
                                                                                                 i]))

    np.savez('/home/bbci/data/results/studies/vitalbci_season1/bbcpy_cv_results_Miklodyfire_%s_%s_excl_%s_lesschans' % (
        estimator, exclstr, str(scoring)),
             cvresultsAlpha=cvresultsAlpha, cvresultsBeta=cvresultsBeta, cvresultsTheta=cvresultsTheta,
             cvresultsBroad=cvresultsBroad)



import matplotlib.pyplot as plt
import numpy as np
import bbcpy
import scipy.io as sio
from vitalbci_funcs import *
import glob
from scipy.stats import ranksums

vp_fb_parameters = sio.loadmat('C:/data/results/studies/vitalbci_season1/vp_fb_parameters.mat')
classes = vp_fb_parameters['classes']

performance = sio.loadmat('C:/data/results/studies/vitalbci_season1/performance.mat')

lwfresults = glob.glob(
    'C:/data/results/studies/vitalbci_season1/bbcpy_cv_results_Miklodyfire_fb_lwf_*_excl_lesschans.npz')
scmresults = glob.glob(
    'C:/data/results/studies/vitalbci_season1/bbcpy_cv_results_Miklodyfire_fb_scm_*excl_lesschan*.npz')

lwfexclparams = [x[x.find('_excl') - 2:x.find('_excl')] for x in lwfresults]
scmexclparams = [x[x.find('_excl') - 2:x.find('_excl')] for x in scmresults]

lwfresults_data = []
for i in range(len(lwfresults)):
    lwfresults_data.append(np.load(lwfresults[i], allow_pickle=True))
scmresults_data = []
for i in range(len(scmresults)):
    scmresults_data.append(np.load(scmresults[i], allow_pickle=True))

dataset = bbcpy.load.eeg.dataset('C:/data/results/studies/vitalbci_season1/bbcpy_datsnip.npy')

bestexcl = np.zeros((len(scmresults_data[0].files), len(scmresults_data[0][scmresults_data[0].files[0]]), 6), int)
bestest = np.zeros((len(scmresults_data[0].files), len(scmresults_data[0][scmresults_data[0].files[0]]), 6), int)
best = np.zeros((len(scmresults_data[0].files), len(scmresults_data[0][scmresults_data[0].files[0]]), 6))
for ifile, file in enumerate(scmresults_data[0].files):
    for isubj in range(len(scmresults_data[0][file])):
        othersubj = np.arange(len(scmresults_data[0][file]))
        othersubj = othersubj[othersubj != isubj]
        res = np.zeros((len(scmresults_data), 2, *scmresults_data[0][file].shape))
        for iexcl in range(len(scmresults_data)):
            res[iexcl, 0] = scmresults_data[iexcl][file]
            res[iexcl, 1] = lwfresults_data[iexcl][file]
        bestest_tmp = res[:, :, :, :, 0][:, :, othersubj].mean(axis=2).argmax(axis=1)
        bestexcl[ifile, isubj] = res[:, :, :, :, 0][:, :, othersubj].mean(axis=2).max(axis=1).argmax(axis=0)
        bestest[ifile, isubj] = bestest_tmp[bestexcl[ifile, isubj], range(len(bestexcl[ifile, isubj]))]
        best[ifile, isubj] = res[bestexcl[ifile, isubj], bestest[ifile, isubj], [isubj] * len(bestexcl[ifile, isubj]),
                                 range(len(bestexcl[ifile, isubj])), 0]
plt.figure()
plt.boxplot(best[0])
ranksums(best[0, :, 1], best[0, :, 3])
plt.figure()
plt.scatter(best[0, :, 1], best[0, :, 3])
plt.plot([0.5, 1], [0.5, 1])

bestexcl_roc = np.zeros((len(scmresults_data[0].files), len(scmresults_data[0][file]), 6), int)
bestest_roc = np.zeros((len(scmresults_data[0].files), len(scmresults_data[0][file]), 6), int)
best_roc = np.zeros((len(scmresults_data[0].files), len(scmresults_data[0][file]), 6))
for ifile, file in enumerate(scmresults_data[0].files):
    for isubj in range(len(scmresults_data[0][file])):
        othersubj = np.arange(len(scmresults_data[0][file]))
        othersubj = othersubj[othersubj != isubj]
        res = np.zeros((len(scmresults_data), 2, *scmresults_data[0][file].shape))
        for iexcl in range(len(scmresults_data)):
            res[iexcl, 0] = scmresults_data[iexcl][file]
            res[iexcl, 1] = lwfresults_data[iexcl][file]
        bestest_tmp = res[:, :, :, :, 1][:, :, othersubj].mean(axis=2).argmax(axis=1)
        bestexcl_roc[ifile, isubj] = res[:, :, :, :, 1][:, :, othersubj].mean(axis=2).max(axis=1).argmax(axis=0)
        bestest_roc[ifile, isubj] = bestest_tmp[bestexcl_roc[ifile, isubj], range(len(bestexcl_roc[ifile, isubj]))]
        best_roc[ifile, isubj] = res[bestexcl[ifile, isubj], bestest_roc[ifile, isubj],
                                     [isubj] * len(bestexcl_roc[ifile, isubj]), range(len(bestexcl[ifile, isubj])), 0]
plt.figure()
plt.boxplot(best_roc[0])
ranksums(best_roc[0, :, 1], best_roc[0, :, 3])



import numpy as np
import pandas as pd
import pyriemann as pr
import sklearn as sk
from sklearn.base import BaseEstimator, TransformerMixin

import bbcpy
from bbcpy.functions.base import ImportFunc
import warnings


def cov(obj, axis=1, estimator='scm', target='trial', keep=True):
    """
    :param obj: obj with dim (nTime x nCh) or (nTime x nCh x nEpo)
    :param axis: -1 for time, -2 (default) for channels, 0 for epochs
    :param estimator: 'SCM' (default) for ... (see doc of pyriemann)
    :param target: 'class' (default), 'all' or 'trial' - subgroups for covariance if trialwise data
    :param keep:
    :return:
    Covariance estimation for continous and epoched data
    """
    if type(obj) == bbcpy.datatypes.eeg.Data:
        # different covariance estimators possible
        if hasattr(obj, '_C') & (axis == -2):  # only makes sense for one possible cov which is mostly ChxCh
            return obj.__C
        C = pr.utils.covariance._check_est(estimator)(np.transpose(obj, (-2, -1)))
        if keep & (axis == -2):
            obj.__C = C
        return C
    elif type(obj) == bbcpy.datatypes.eeg.Epo:
        # different covariance estimators possible
        if hasattr(obj, '_C') & (axis == -2) & (target == 'class'):  # only makes sense if only one possible cov
            return obj.__C
        if np.ndim(obj) == 3:
            if (axis == 2) | (axis == -1):
                permord = (2, 1, 0)
                reshapeord = [obj.nT(), obj.nCh()]
                targetint = 1
                covdim = obj.shape[axis]
            elif (axis == 1) | (axis == -2):
                permord = (1, 2, 0)
                reshapeord = [obj.nCh(), obj.nT()]
                targetint = 1
                covdim = obj.shape[axis]
            elif axis == 0:
                permord = (0, 2, 1)
                reshapeord = [1, obj.nT() * obj.nCh()]
                targetint = 0
                if target == 'class':
                    covdim = obj.nClass()
                else:
                    covdim = obj.shape[axis]
                if target == 'trial':
                    warnings.warn('.cov() Ignoring target \'trial\' for axis = 2 because cannot calculate covariance'
                                  ' over trials for each trial')
                    target = 'all'
            else:
                raise ValueError('Unknown axis for cov(): ' + str(axis))
        else:
            if (axis == 1) | (axis == -1):
                permord = (1, 0)
                reshapeord = [obj.shape[1], 1]
                targetint = 1
                covdim = obj.shape[axis]
            elif axis == 0:
                permord = (0, 1)
                reshapeord = [1, obj.shape[1]]
                targetint = 0
                if target == 'class':
                    covdim = obj.nClass()
                else:
                    covdim = obj.shape[axis]
                if target == 'trial':
                    warnings.warn('.cov() Ignoring target \'trial\' for axis = 2 because cannot calculate covariance'
                                  ' over trials for each trial')
                    target = 'all'
            else:
                raise ValueError('Unknown axis for cov(): ' + str(axis))
        if target == 'class':
            C = np.empty((obj.nClass(), covdim, covdim))
            reshapeord2 = reshapeord.copy()
            for i in range(obj.nClass()):
                reshapeord2[targetint] = reshapeord[targetint] * (obj.y == i).sum()
                epoc = np.transpose(obj[obj.y == i], permord)
                epoc = epoc.reshape(reshapeord2)
                C[i] = pr.utils.covariance._check_est(estimator)(epoc)
        elif target == 'trial':
            C = np.empty((obj.nEpo(), covdim, covdim))
            for i in range(obj.nEpo()):
                epoc = np.transpose(obj[i], permord)
                epoc = epoc.reshape(reshapeord)
                C[i] = pr.utils.covariance._check_est(estimator)(epoc)
        elif target == 'all':
            reshapeord[targetint] *= obj.nEpo()
            C = np.empty((covdim, covdim))
            epoc = np.transpose(obj, permord)
            epoc = epoc.reshape(reshapeord)
            C = pr.utils.covariance._check_est(estimator)(epoc)
        else:
            raise ValueError('Unknown target for cov(): ' + target)
        if keep & (axis == -2) & (target == 'class'):
            obj.__C = C
        return C




class ZScore(BaseEstimator, TransformerMixin):
    """Calculate z-score.

        Parameter
        ----------
        mean : float, optional (default: None)
            mean to use for zscoring
        std : float, optional (default: None)
            standard deviation used for zscoring
        online : Bool, optional (default: False)
            trigger for Online zscore estimation. If True, zscore is estimated
            in windows with lenght specified in the 'window' parameter.
        window : int, optional (default: 1000)
            Window length for online zscore estimation. Ignored if
            online=False.
    """

    def __init__(self, mean=None, std=None, online=False, window=1000):
        self.online = online
        self.mean = mean
        self.std = std
        self.window = window

    def fit(self, X, y=None):
        """Fit the Transformer.

        Parameter
        ----------
        X : ndarray, shape (n_samples, n_channels)
            Training data.
        y : ndarray, shape (n_samples), optional (default: None)
            Training data labels.

        """

        if not self.online:
            if self.mean is None:
                self.mean = np.mean(X, axis=0)
            if self.std is None:
                self.std = np.std(X, axis=0)

        return self

    def transform(self, X, y=None):
        """Fit the Transformer.

        Parameter
        ----------
        X : ndarray, shape (n_samples, n_channels)
            Data set.
        y : ndarray, shape (n_samples), optional (default: None)
            Labels for data set.

        """
        if self.online:
            m = np.array([pd.Series(xtmp).rolling(window=self.window,
                                                  min_periods=1).mean() for xtmp
                          in X.T]).T
            s = np.array([pd.Series(xtmp).rolling(window=self.window,
                                                  min_periods=1).std() for xtmp
                          in X.T]).T
            s[0] = np.ones(X.shape[-1])

            zs = (X - m) / s
        else:
            zs = (X - self.mean) / self.std
        # nan_to_num is needed when a signal is constant over a window length
        return np.nan_to_num(zs, posinf=0.0, neginf=0.0)




#todo re-write as simple function?
class LogVariance(sk.base.BaseEstimator, sk.base.TransformerMixin):
    """Return the logarithm of the variance
    """

    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return np.log(np.var(X, axis=2))

# Implementing simple numpy var as a transformer type:
var = ImportFunc(np.var, axis=2)

#todo delete the following?
class Variance(sk.base.BaseEstimator, sk.base.TransformerMixin):
    """Return the logarithm of the variance
    """

    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return np.var(X, axis=2)
